
/*
 *                      -- bmr.c --
 *           (Boot Manager Recovery program)
 *
 *   Copyright (c) 2001 Efim Tokarevski
 *
 * You may distribute this file under the terms of the GNU General Public
 * License.  See the file COPYING for more information.
 *
 *
 *   This file has been compiled linked and debugged by:
 *
 *   Microsoft (R) C Optimizing Compiler Version 6.00A
 *   Copyright (c) Microsoft Corp 1984-1990. All rights reserved.
 * 
 *   Microsoft (R) Segmented-Executable Linker  Version 5.10 
 *   Copyright (C) Microsoft Corp 1984-1990.                 
 *                                                           
 *   Microsoft (R) CodeView (R)  Version 3.00                
 *   Copyright (C) Microsoft Corp 1986-1990.                 
 * 
 * 
 */

#pragma check_pointer   (off)       /* otherwise - run-time error */
                                    /* while reading BIOS data    */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include <conio.h>
#include <dos.h>
#include <malloc.h>

#define CHS_valid       0x02        /* see: dparams.flags */

char    _far *fixdrives = (char _far *)0x00400075;/* 0040:0075 disk drives num */

jmp_buf         mark;
FILE           *fptr;
char           *progname;
unsigned int    fcount;             /* processed files counter */
unsigned int    vrb_flag;
char            mode_flag = -1;     /*  0 -CHS; 0x40 -EDD */
int             return_value;

typedef struct mboot{
    unsigned char mbr_body[0x1b6];
    unsigned long lba;
    unsigned long ver;
    unsigned char ptab[0x40];       /* part table must be at 0x1be */
    unsigned int mark;
} *Mboot;

void flags_proc (char *token);
void fname_proc (char *filename);
char answer (char);

char legend[] = "\n\
Syntax:   %s [<opts>] {<filename>}\n\
Function: restores Master Boot Record(s) from files\n\
          backuped by 'BMI.EXE' ( Boot Manager Installation program)\n\
Options:\n\
     -e     force EDD mode\n\
     -c     force CHS mode\n\n\
filename is supposed to have the form of:  ffffffff.nnn\n\n\
    where:              ffffffff    - lba of an MBR (in hex code)\n\
                        nnn         - hard disk number (001-...)\n\n\
";


main (int argc, char *argv[], char *envp[]) {
    int  i;
    char *tmp;

    return_value = setjmp(mark);
    if (return_value < 0) {
        if (fptr) fclose (fptr);
        if (return_value == -1) {
            if (errno) perror (NULL);
        }
        return (return_value);
    }

    progname = argv[0];
    while ((tmp = strpbrk (progname, ":\\")) != NULL) progname = ++tmp;

    if (argc == 1) printf (legend, progname);
    else {
        for (i=1; argv[i]; i++) flags_proc (argv[i]);
        for (i=1; argv[i]; i++) fname_proc (argv[i]);
        if (!fcount) printf ("Nothing to do\n");
    }
    return (return_value);
}


void fname_proc (char *filename) {
    Mboot           buffer_ptr;
    char _far      *buffer_fptr;
    int             modflag;
    unsigned int    fsize;
    unsigned int    blcnt;
    unsigned int    disk;
    unsigned char  *disk_ptr;
    unsigned long   lba;
    unsigned char  *lba_ptr;
    unsigned long   tmp;
    unsigned char  *tmp_ptr;
    unsigned long   cyl;
    unsigned long   head;
    unsigned long   sect;

    union REGS      inp_regs;
    union REGS      outp_regs;
    struct SREGS    segregs;

    struct {
        unsigned int    pktsize;    /* (call) size of buffer, (ret) size of ret data    */
        unsigned int    flags;      /* information flags                                */
        unsigned long   ncyls;      /* number of physical cylinders on drive            */
        unsigned long   nheads;     /* number of physical heads on drive                */
        unsigned long   nsecs;      /* number of physical sectors per track             */
        unsigned long   totsecs[2]; /* total number of sectors on drive                 */
        unsigned int    sectsize;   /* bytes per sector                                 */
    }dparams = {sizeof dparams};
    char _far *p_dparams =(char _far *)&dparams;

    struct {
        unsigned char   pktsize;
        unsigned char   res;
        unsigned int    blocks;
        unsigned char   _far *buff;
        unsigned long   lba[2];
    }pkt = {sizeof pkt,0};
    char _far *p_pkt = (char _far *)&pkt;


    if (*filename == '-' || *filename == '/') return;
    fcount++;

    if (fptr) fclose (fptr);    /* open one file at a time */
    if ((fptr = fopen (filename, "rb+")) == NULL) {
        if (errno == 0) {
            printf ("%s: fopen error\n", filename);
            longjmp(mark, -2);
        }
        else longjmp(mark, -1);
    }
    if (fseek (fptr, 0L, SEEK_END))  longjmp(mark, -1);

    fsize = ftell (fptr);
    if (fsize == -1L)  longjmp(mark, -1);

    if (fseek (fptr, 0L, SEEK_SET))  longjmp(mark, -1); /* beg of file*/

    lba_ptr = filename;
    while ((tmp_ptr = strpbrk (lba_ptr, ":\\")) != NULL)
                                         lba_ptr = ++tmp_ptr;

    if ((tmp_ptr = strpbrk (lba_ptr, ".")) == NULL) {
        printf ("%s: can't define MBR address from this filename\n", filename);
        longjmp(mark, -2);
    }

    tmp = 0;
    *tmp_ptr = 0;
    sscanf (lba_ptr, "%8lx%n", &lba, &tmp);
    if (tmp != 8) {
        printf ("%s: can't define MBR address from this filename\n", filename);
        longjmp(mark, -2);
    }
    *tmp_ptr = '.';

    tmp = 0;
    disk_ptr = ++tmp_ptr;
    sscanf (disk_ptr, "%3d%n", &disk, &tmp);
    if (tmp != 3 || disk > *fixdrives || disk == 0) {
        printf ("%s: can't define disk drive id from this filename\n", filename);
        longjmp(mark, -2);
    }

    if (mode_flag < 0) {
        modflag = 0;
        inp_regs.h.ah = 0x41;               /* EDD installation check */
        inp_regs.h.dl = (disk - 1) | 0x80;
        inp_regs.x.bx = 0x55aa;
        int86x (0x13, &inp_regs, &outp_regs, &segregs);
        if (outp_regs.x.cflag == 0 &&
            outp_regs.x.bx == 0xaa55 &&
            (outp_regs.x.cx & 1) == 1) modflag = 0x40;
    }
    else modflag = mode_flag;

    inp_regs.h.ah = 0x08 | modflag;         /* get drive parameters */
    inp_regs.h.dl = (disk - 1) | 0x80;
    inp_regs.x.si = FP_OFF(p_dparams);
    segregs.ds    = FP_SEG(p_dparams);
    int86x (0x13, &inp_regs, &outp_regs, &segregs);
    if (outp_regs.x.cflag) {
        printf ("\nerror while getting drive #%d parameters, "
                "error code (ah): %#04hx\n", disk, outp_regs.h.ah);
        longjmp(mark, -2);
    }

    if (modflag == 0) {
        dparams.ncyls = ((outp_regs.h.cl & 0xc0) * 4) + outp_regs.h.ch + 1;
        dparams.nheads = outp_regs.h.dh + 1;
        dparams.nsecs = (outp_regs.h.cl & 0x3f);
        dparams.totsecs[0] = dparams.ncyls * dparams.nheads * dparams.nsecs;
        dparams.flags = CHS_valid;
        dparams.sectsize = 512;             /* 512 -assumed for CHS mode */
    }

    if (vrb_flag) {
        printf ("---------------\n");
        printf ("disk number: %d\n", disk);
        if (dparams.flags & CHS_valid) {
            printf ("number of cylinders on the drive: %ld\n", dparams.ncyls);
            printf ("number of tracks per one cylinder: %ld\n", dparams.nheads);
            printf ("number of sectors per one track: %ld\n", dparams.nsecs);
        }
        printf ("number of bytes per one sector: %d\n", dparams.sectsize);
        printf ("total number of sectors on the drive: %ld\n", dparams.totsecs[0]);
    }

    if (lba >= dparams.totsecs[0]) {
        printf ("lba = %ld; this is greater than max lba value for the disk #%d\n", lba, disk);
        longjmp(mark, -2);
    }

    if (modflag == 0) {
        cyl = lba / (dparams.nsecs * dparams.nheads ); /* cyl:  0 - ... */
        tmp = lba % (dparams.nsecs * dparams.nheads);
        head = tmp / dparams.nsecs;                    /* head: 0 - ... */
        sect = tmp % dparams.nsecs + 1;                /* sect: 1 - ... */
    }

    blcnt = fsize / dparams.sectsize;
    if (
         blcnt == 0                 ||
         blcnt > 2                  ||
         (blcnt == 2 && lba !=0)
        ) {
        printf ("%s: size of this file isn't valid for MBR backup file\n", filename);
        longjmp(mark, -2);
    }

    if (vrb_flag) {
        printf ("\n");
        printf ("number of sectors to transfer: %d\n", blcnt);
        printf ("LBA of the first sector:  %ld\n", lba);
        if (modflag == 0) {
            printf ("cylinder: %ld\n", cyl);
            printf ("head: %ld\n", head);
            printf ("sector: %ld\n", sect);
        }
    }

    if ((buffer_ptr = malloc (dparams.sectsize * 2)) == NULL) {
        if (errno == 0) {
            printf ("malloc error\n");
            longjmp(mark, -2);
        }
        else longjmp(mark, -1);
    }
    buffer_fptr = (char _far *)buffer_ptr;
    buffer_ptr->lba = 0xffffffff;

    inp_regs.h.ah = 0x02 | modflag;         /* disk read */
    inp_regs.h.dl = (disk - 1) | 0x80;
    if (modflag == 0) {
        inp_regs.h.al = blcnt;
        inp_regs.h.ch = cyl & 0xff;
        inp_regs.h.cl = sect + (cyl >> 2 & 0xc0);
        inp_regs.h.dh = head;
        inp_regs.x.bx = FP_OFF(buffer_fptr);
        segregs.es    = FP_SEG(buffer_fptr);
    }
    else {
        inp_regs.x.si = FP_OFF(p_pkt);
        segregs.ds    = FP_SEG(p_pkt);
        pkt.blocks = blcnt;
        pkt.lba[0] = lba;
        pkt.lba[1] = 0;
        pkt.buff = buffer_fptr;
    }
    int86x (0x13, &inp_regs, &outp_regs, &segregs);
    if (outp_regs.x.cflag) {
        printf ("\nerror while reading from disk #%d, "
                "error code (ah): %#04hx\n", disk, outp_regs.h.ah);
        longjmp(mark, -2);
    }

    if (buffer_ptr->lba != lba || buffer_ptr->mark != 0xaa55)
        printf ("\nCaution: the sector you are trying to replace with %s\n"
                  "         is not a valid MBR sector!\n", filename);

    if ((fread (buffer_ptr, dparams.sectsize, blcnt, fptr)) != blcnt) {
        if (errno == 0) {
            printf ("%s: fread error\n", filename);
            longjmp(mark, -2);
        }
        else longjmp(mark, -1);
    }

    printf ("filename:  %s y/n/q", filename);
    if ((answer ('y') | 0x20) != 'y') {
        printf ("\n");
        return;
    }

    inp_regs.h.ah = 0x03 | modflag;         /* disk write */
    inp_regs.h.dl = (disk - 1) | 0x80;
    if (modflag == 0) {
        inp_regs.h.al = blcnt;
        inp_regs.h.ch = cyl & 0xff;
        inp_regs.h.cl = sect + (cyl >> 2 & 0xc0);
        inp_regs.h.dh = head;
        inp_regs.x.bx = FP_OFF(buffer_fptr);
        segregs.es    = FP_SEG(buffer_fptr);
    }
    else {
        inp_regs.h.al = 0; /* flags */
        inp_regs.x.si = FP_OFF(p_pkt);
        segregs.ds    = FP_SEG(p_pkt);
        pkt.blocks = blcnt;
        pkt.lba[0] = lba;
        pkt.lba[1] = 0;
        pkt.buff = buffer_fptr;
    }
    int86x (0x13, &inp_regs, &outp_regs, &segregs);
    if (outp_regs.x.cflag) {
        printf ("\nerror while writing to disk #%d, "
                "error code (ah): %#04hx\n", disk, outp_regs.h.ah);
        longjmp(mark, -2);
    }
    printf (" ok.\n");
}


void flags_proc (char *token) {
    if (*token == '-' || *token == '/') {
        token++;
        switch (*token) {
            case 'e': mode_flag = 0x40;
                      break;
            case 'c': mode_flag = 0;
                      break;
            case 'v': vrb_flag = 1;
                      break;
            default : fprintf (stderr, "No such option '%s'\n", token);
                      printf (legend, progname);
                      longjmp(mark, -2);
        }
    }
    return;
}


char answer (char defchar) {
    char tmp;

    printf (" [%c]", defchar);
    tmp = (char) getch();
    if ((tmp | 0x20) == 'q') longjmp(mark, -2);
    if (tmp == 0x0d) tmp = defchar;
    return (tmp);
  }
